#include <map>
#include <string>
#include <vector>

#include <stringlib.h>
#include "general.h"
#include "type_value.h"
#include "flowctrl.h"


typedef type_value &(*TemplateParserLookupVarFnc)(char const *,
                                                  bool &,
                                                  void *,
                                                  void *);
typedef flowctrl (*TemplateParserExecFuncFnc)(string,
                                              void *,
                                              void *,
                                              joined_list<type_value> &);

struct FunctionExpr {
    string functionName;
    joined_list<type_value> args;
    int argCount;
    
    FunctionExpr(string name) 
    :   functionName(name), argCount(0) {
    }
    
    void addArg(type_value &arg) {
        args.insert_listelem(arg, argCount++);
    }
};

class Command
{
public:
    typedef bool (*CmdFnc1)(std::string const &);
    typedef bool (*CmdFnc2)(std::string const &, std::string const &);

    Command()
    {
    }

    Command(CmdFnc1 cmdFnc1)
    :   argc_(1), cmdFnc1_(cmdFnc1)
    {
    }

    Command(CmdFnc2 cmdFnc2)
    :   argc_(2), cmdFnc2_(cmdFnc2)
    {
    }

    bool Execute(std::string const &arg1) const
    {
        return cmdFnc1_(arg1);
    }

    bool Execute(std::string const &arg1,
                 std::string const &arg2) const
    {
        return cmdFnc2_(arg1, arg2);
    }
    
    int ArgCount() const
    {
        return argc_;
    }

    bool operator==(CmdFnc1 fnc) const
    {
        return cmdFnc1_ == fnc;
    }

private:
    int argc_;
    CmdFnc1 cmdFnc1_;
    CmdFnc2 cmdFnc2_;
};

class TemplateParser
{
public:
    TemplateParser(TemplateParserLookupVarFnc lookupFnc, 
                   TemplateParserExecFuncFnc execFnc,
                   void *lookupEnv1, void *lookupEnv2);

    void Parse();
    ::string GetResult(bool &error_occured,
                       bool &flow_error_occured,
                       std::string &error_msg);
    ::flowctrl GetFlowError();
    void SetSrc(const ::string &src);
    
private:
    
    struct CharRange {
        char const *begin, *end;
        
        CharRange(char const *b, char const *e)
        :   begin(b), end(e) {
        }
        
        bool empty() const {
            return end == begin;
        }
        
        int size() const {
            return end - begin;
        }
        
        char operator[](size_t pos) const {
            return begin[pos];
        }
    };
    
    typedef std::map<std::string, Command> CommandMap;
    typedef CommandMap::iterator CommandIterator;

    TemplateParserLookupVarFnc lookupFnc_;
    TemplateParserExecFuncFnc execFnc_;

    void *lookupEnv1_, *lookupEnv2_;

    bool Eof() const
    {
        return srcPtr_ == srcEnd_;
    }

    bool BraceAhead();
    bool ElseAhead();

    void Error(char const *msg);
    void Error(CharRange const &charRange, char const *msg) {
        srcPtr_ = charRange.begin;
        Error(msg);
    }

    Command const &GetCommand(std::string const &cmdName);
    char const *FindNextPipe(char const *p);
    char const *FindNextPipeOrParen(char const *p);
    char const *FindNextPipeOrSharp(char const *p);

    char const *LookupVar(char const *var, bool &found);
    type_value &LookupVarValue(char const *var, bool &found);
    type_value ExecFunc(FunctionExpr &funcExpr);

    void ParseBody(bool skip);
    void ParseCommand(bool skip);
    std::string ParseCommandName();
    bool ParseDoubleSharp();
    void ParseElse(bool skip);
    std::string ExecuteFunctionExprToStr(CharRange const &expr,
                                         bool &found);
    type_value ExecuteFunctionExpr(CharRange &expr,
                                   bool &found);
    std::string ParseFirstCommandArgument(bool skip, bool &found);
    std::string ParseSecondCommandArgument(bool skip);
    void ParseLiteralText(bool skip); // May be empty, if eof.
    void ParseRange(bool skip);
    void ParseReplaceExpr(bool skip);
    void ParseReplaceVar(bool skip);
    void ParseText(bool skip); // May be empty, if eof.

    bool ParseORExpression(bool skip);
    bool ParseANDExpression(bool skip);
    bool ParseNOTExpression(bool skip);
    bool ParseSimpleExpression(bool skip);
    bool ParseOperator(char const *op);
    void SkipWhite();
    static void SkipWs(CharRange &s);
    CharRange ParseName(CharRange &s);
    void ParseFunctionArg(CharRange &args, FunctionExpr &fexpr);
    void ParseFunctionArgs(CharRange &args, FunctionExpr &fexpr);
    bool ParseLiteral(CharRange &expr, type_value &literalValue);
    bool ParseStringLiteral(CharRange &expr, type_value &literalValue);
    bool ParseLongLiteral(CharRange &expr, type_value &literalValue);
    bool ParseDoubleLiteral(CharRange &expr, type_value &literalValue);
    int HexValue(CharRange const &charRange, char c);

    void Write(char const *str, size_t strLen);
    void Write(char const *strBegin, char const *strEnd);
    void Write(char const *str);

    ::string src_;
    ::string dest_;

    char const *srcPtr_, *srcBegin_, *srcEnd_;

    bool        error_occured_, flow_error_occured_;
    std::string error_msg_;

    flowctrl flowError_;

    int depth_;
    static CommandMap cmds_;
    
};

inline void TemplateParser::Write(char const *str, size_t strLen)
{
    dest_.append(str, strLen);
}
